﻿#include "TCP_client.h"
UINT16 Transmission_Indicator = 0x00; //事务号

/*********************************************************************************************
* 功能    ：	打印读取从站的线圈状态
* 描述	  ：	对响应报文中的数据进行解析和显示
* 输入    ：    Response_Message 响应报文  Request_Message 请求报文
* 输出    ：    无
*********************************************************************************************/
void Printf_Coil_date(UINT8 *Response_Message, UINT8 *Request_Message)
{
	printf("从站设备ID %02X 功能码为 %02X\n", Response_Message[6], Response_Message[7]);
	unsigned int temp1 = ((Request_Message[8] << 8) | Request_Message[9]);
	unsigned int number = ((Request_Message[10] << 8) | Request_Message[11]);
	printf("线圈起始地址为%d \n",temp1);
	unsigned int temp = temp1;
	for (int i = 0; i < Response_Message[8]; i++)
	{
		unsigned int temp2 = temp + 7;
		if (temp2 > temp1 + number - 1)
			temp2 = temp1 + number - 1;
		printf("线圈%d --- %d的状态为：%02X \n", temp2, temp, Response_Message[9+i]);
		temp = temp + 8;
	}
}

/*********************************************************************************************
* 功能    ：	打印读取从站的寄存器状态
* 描述	  ：	对响应报文中的数据进行解析和显示
* 输入    ：    Response_Message 响应报文  Request_Message 请求报文
* 输出    ：    无
*********************************************************************************************/
void Printf_Register_date(UINT8 *Response_Message, UINT8 *Request_Message)
{
	printf("从站设备ID %02X 功能码为 %02X\n", Response_Message[6], Response_Message[7]);
	unsigned int temp1 = ((Request_Message[8] << 8) | Request_Message[9]);
	printf("寄存器起始地址为%d \n", temp1);
	for (int i = 0; i < Response_Message[8]; i = i + 2)
	{
		printf("寄存器%d的值为：%02X %02X \n", temp1++, Response_Message[9 + i], Response_Message[10+i]);
	}
}

/*********************************************************************************************
* 功能    ：	打印读取异常响应报文信息
* 描述	  ：	对响应报文中的数据进行解析和显示
* 输入    ：    Response_Message 响应报文
* 输出    ：    无
*********************************************************************************************/
void Printf_Anomaly_date(UINT8 *Response_Message)
{
	UINT8 anomaly_code = Response_Message[8];
	printf("从站设备ID %02X 功能码为 %02X\n", Response_Message[6], Response_Message[7]);
	switch (anomaly_code)
	{
	case 0x01:	printf("%02X : 从站设备不支持此功能码\n", anomaly_code); break;
	case 0x02:	printf("%02X : 指定的数据地址在从站设备中不存在\n", anomaly_code); break;
	case 0x03:	printf("%02X : 指定的数据超过范围或者不允许使用\n", anomaly_code); break;
	case 0x04:	printf("%02X : 从站设备处理响应的过程中，出现未知错误等\n", anomaly_code); break;
	default:	printf("Unkown Other Error!!!!!\n");
	}
}

/*********************************************************************************************
* 功能    ：	检测响应报文长度
* 描述	  ：	对响应报文中存放长度字节进行判断和请求报文对比
* 输入    ：    Response_Message 响应报文 Request_Message 请求报文 Response_Message_len 接收到的数据长度
* 输出    ：    true 长度正常  false 长度异常
*********************************************************************************************/
bool Check_Response_Message_len(UINT8 *Response_Message, UINT8 *Request_Message, int response_message_len)
{
	if ((response_message_len - 6) != Response_Message[5])
		return false;
	unsigned int operations_Number = Request_Message[10] << 8 | Request_Message[11];
	unsigned int respone_Len = Count_Respone_Len(Request_Message[7], operations_Number);
	if (Request_Message[7] == 0x0F || Request_Message[7] == 0x10)
		if (response_message_len != 12)
			return false;
	if (Request_Message[7] == 0x01 || Request_Message[7] == 0x03)
	{
		if (respone_Len + 9 != response_message_len)
			return false;
		if (respone_Len != Response_Message[8])
			return false;
	}
	return true;
}

bool Printf_Analysis_Outcome(UINT8 *Response_Message, UINT8 *Request_Message)
{
	switch (Response_Message[7])
	{
	case 0x01:Printf_Coil_date(Response_Message, Request_Message); break;
	case 0x03:Printf_Register_date(Response_Message, Request_Message); break;
	case 0x0F:printf("成功写入从站线圈%d个\n", ((Request_Message[10] << 8) | Request_Message[11])); break;
	case 0x10:printf("成功写入从站寄存器%d个\n", ((Request_Message[10] << 8) | Request_Message[11])); break;
	default: return false;
	}
	return true;
}
/*********************************************************************************************
* 功能    ：	判断响应报文是否可以正常解析
* 描述	  ：	通过长度和异常码等判定该响应报文是否可以解析
* 输入    ：    Response_Message 响应报文 Request_Message请求报文  Response_Message_len 响应报文长度
* 输出    ：    true可以正常解析 false 无法解析
*********************************************************************************************/
bool Analysis_Response_Message(UINT8 *Response_Message, UINT8 *Request_Message, int response_message_len)
{
	if (response_message_len < 9 || response_message_len > 260)
		return false;
	for (int i = 0; i < 7; i++)
	{
		if (i == 4 || i == 5) //后续字节长度
			continue;
		if (Response_Message[i] != Request_Message[i]) //0 1 传输标识，2 3 协议标识，6设备ID 
			return false;  //一旦不一致，判定为异常报文数据
	}
	if (Response_Message[4] != 0x00) //4 固定0x00
		return false;

	if (Response_Message[7] == Request_Message[7] + 0x80)//先处理异常响应
	{
		if (response_message_len == 9)
		{
			Printf_Anomaly_date(Response_Message);
			return true;
		}
		else
			return false;
	}
	if (Response_Message[7] != Request_Message[7]) //功能码判断
		return false;
	if (Request_Message[7] == 0x0F || Request_Message[7] == 0x10) //0F 10 地址和操作数量判定
	{
		for (int j = 8; j < 12; j++)
		{
			if (Request_Message[j] != Response_Message[j])
				return false;
		}
	}
	if (!Check_Response_Message_len(Response_Message, Request_Message, response_message_len))
		return false;
	//打印解析结果
	return Printf_Analysis_Outcome(Response_Message, Request_Message);
}


void Printf_Message(UINT8 *Message,int flage, int message_len)
{
	if (flage == 1)
		printf("主站请求 ：");
	else
		printf("从站响应 ：");
	for (int i = 0; i < message_len; i++)
	{
		printf("%02x  ", Message[i]);
	}
	printf("\n");
	if (LOG_NOTE_SWITCH)
		Log_Note(Message, flage, message_len);
}


/*********************************************************************************************
* 功能    ：	生成MBAP报头
* 描述    ： MBAP报文头的包括的内容：
* +-------------+---------+--------+--------+------------------------------+
* |     域      |  长度   | 客户机 | 服务器 |              描述            |
* +-------------+---------+--------+--------+------------------------------+
* |事务元标识符 | 2个字节 |  启动  |  复制  |请求/响应事务处理的识别码     |
* +-------------+---------+--------+--------+------------------------------+
* |协议标识符   | 2个字节 |  启动  |  复制  |0=MODBUS 协议                 |
* +-------------+---------+--------+--------+------------------------------+
* |    长度     | 2个字节 |  启动  |  启动  |以下字节的数量                |
* +-------------+---------+--------+--------+------------------------------+
* |单元标识符   | 1个字节 |  启动  |  复制  |连接的远程从站的识别码        |
* 输入    ：    无
* 输出    ：    MBAP报文头内容
*********************************************************************************************/
void Create_MBAP(UINT8 *Message, int function_code, unsigned int operations_number)
{
	Message[0] = Transmission_Indicator >> 8;//事务号
	Message[1] = (UINT8)Transmission_Indicator;
	Transmission_Indicator++;
	Message[2] = 0x00; //Modbus协议标识
	Message[3] = 0x00;
	Message[4] = 0x00;//后续字节长度
	Message[5] = Count_Write_date_number(function_code, operations_number) + 0x06;
	if (function_code == 0x0F || function_code == 0x10)//0f/10功能码后续字节数多1
	{
		Message[5] = Message[5] + 1;
	}
	Message[6] = DEVICE_ID;
}

/*********************************************************************************************
* 功能    ：	字符串转UINT8类型
* 描述	  ：	根据写入数量和功能码类型输入相应的数据
* 输入    ：    *Message 消息帧存放的数组   Write_date 要转换的字符串
*				Message_len消息帧数组的起始位置
* 输出    ：    Message_len 消息帧的长度
*********************************************************************************************/
int HexStringtoByte(UINT8 *Message, string write_date, int message_len, int function_code, unsigned int operations_Number)
{
	if (write_date.length() == 0)
		return message_len;
	Message[message_len] = Count_Write_date_number(function_code, operations_Number);
	message_len++;
	const char *b = write_date.c_str();
	int temp = 0;
	for (unsigned int i = 0; i + 3 < write_date.length(); i = i + 3)
	{
		sscanf_s(b + i, "%02X", &temp);
		Message[message_len] = (UINT8)temp;
		message_len++;
	}
	sscanf_s(b + write_date.length() - 2, "%02X", &temp);
	Message[message_len] = (UINT8)temp;
	return ++message_len;
}

/*********************************************************************************************
* 功能    ：	生成TCP模式下的消息帧
* 描述	  ：	根据要写入的数据生成消息帧
* 输入    ：    *Message 消息帧存放的数组   Write_date 写入数据的字符串
*				Function_code 功能码   Operations_Number 操作数量  Starting_address起始地址
* 输出    ：    消息帧的总长度
*********************************************************************************************/
int Create_TCP_Message(UINT8 *Message, int function_code, unsigned int operations_Number, unsigned int starting_address, string write_date)
{

	Create_MBAP(Message, function_code, operations_Number);
	Message[7] = function_code;//功能码
	Message[8] = starting_address >> 8;//起始地址H
	Message[9] = starting_address;//起始地址L
	Message[10] = operations_Number >> 8;//操作数量H
	Message[11] = operations_Number;//操作数量L

	return HexStringtoByte(Message, write_date, 12, function_code, operations_Number);

}

/*********************************************************************************************
* 功能    ：	运行客户端
* 描述	  ：	初始化并且生成发送请求等待响应
* 输入    ：    无
* 输出    ：    无
*********************************************************************************************/
int Tcp_Client()
{
	SOCKET clientSocket;
	UINT8 Request_Message[600];
	memset(Request_Message, 0, 600);
	UINT8 Response_Message[600];
	memset(Response_Message, 0, 600);

	do
	{
		clientSocket = Init_Client();
	} while (INVALID_SOCKET == clientSocket);

	while (true)
	{
		int function_code = Input_Function_code();//交互
		unsigned int operations_Number = Input_Operations_number(function_code);
		unsigned int starting_address = Input_Starting_address();
		string write_date = Input_Write_date(function_code, operations_Number);
		int request_message_len = Create_TCP_Message(Request_Message, function_code, operations_Number, starting_address, write_date);//生成
		if (!Send_date(clientSocket, Request_Message, request_message_len)) //发送
		{
			if (Abnormal_Connection(&clientSocket))
				continue;
			else
				break;
		}
		int Response_Message_len = Recv_date(clientSocket, Response_Message);//接收
		Printf_Message(Request_Message, 1, request_message_len);
		Printf_Message(Response_Message, 0, Response_Message_len);
		if (Response_Message_len)
		{
			if (!Analysis_Response_Message(Response_Message, Request_Message, Response_Message_len))
				printf("响应报文数据异常\n");
		}
		else if(!Test_Connection_status(clientSocket))
		{
			if (Abnormal_Connection(&clientSocket))
				continue;
			else
				break;
		}
		else
			printf("响应超时\n");

		printf("**************************** Press Enter To Contioun ****************************\n");
		getchar();
		system("cls");
	}
	//关闭套接字
	closesocket(clientSocket);
	//关闭服务
	WSACleanup();
	return 0;
}
