﻿#include "common.h"


/*********************************************************************************************
* 功能    ：	IP地址有效性检测
* 描述	  ：	检测输入的IP地址是否合法
* 输入    ：    IP   输入的IP地址
* 输出    ：    true  IP地址合法
*				false IP地址非法
*********************************************************************************************/
bool Check_IP(string ip)
{
	int s[4];
	if (ip.length() < 7 || ip.length() > 15) //长度判定
		return false;
	if (sscanf_s(ip.c_str(), "%d.%d.%d.%d", &s[0], &s[1], &s[2], &s[3]) != 4) //IPV4格式正确
	{
		return false;
	}
	string newip = to_string(s[0]) + "." + to_string(s[1]) + "." + to_string(s[2]) + "." + to_string(s[3]);
	if (ip != newip) //前导0
		return false;
	if ((s[0] & 0xffffff00) || (s[1] & 0xffffff00) || (s[2] & 0xffffff00) || (s[3] & 0xffffff00)) //判断每一段大小是否符合要求
	{
		return false;
	}
	return true;
}

/*********************************************************************************************
* 功能    ：	获取从站IP地址和端口号
* 描述	  ：	终端输入从站IP地址和端口号
* 输入    ：    IP 地址   （IPV4）
*				*Port_number 端口号（1-65535）
* 输出    ：    无
*********************************************************************************************/
void Input_IP(string& ip, unsigned int *port_number)
{
	int i = 1;
	do
	{
		if (i == 1)
		{
			cout << "请输入从站IP：";
			i = 0;
		}
		else
			cout << "IP地址格式不正确，请重新输入从站IP：";
		cin >> ip;
	} while (!Check_IP(ip));

	i = 1;
	cout << "IP地址输格式入正确，请输入从站端口号：";
	do
	{
		if (i != 1)
			cout << "请重新输入从站端口号：";
		cin >> *port_number;
		i = 0;
	} while (*port_number == 0 || *port_number > 65535); //端口不能为0 ，端口号范围1---65535

}

/*********************************************************************************************
* 功能    ：	选择功能码
* 描述	  ：	在功能码 0x01 0x03 0x0F 0x10中选择一个功能码
* 输入    ：    无
* 输出    ：    Function_code  选择的功能码
*********************************************************************************************/
int Input_Function_code(void)
{
	int function_code;
	cin >> hex;
	int i = 1;
	printf("功能码 ：  0x01      0x03        0x0F         0x10 \n");
	printf("说明   ： 读线圈   读寄存器   写多个线圈  写多个寄存器\n");
	do
	{
		if (i == 1)
		{
			i = 0;
			cout << "请输入选择的功能码：";
		}
		else
			cout << "输入功能码错误，请重新选择的功能码";
		cin >> function_code;
		cin.clear();
		cin.sync();
	} while (function_code != 0x01 && function_code != 0x03 && function_code != 0x0F && function_code != 0x10);
	return function_code;
}

/*********************************************************************************************
* 功能    ：	输入操作起始地址码
* 描述	  ：	起始地址码，最大值65535
* 输入    ：    无
* 输出    ：    starting_address 操作起始地址码
*********************************************************************************************/
unsigned int Input_Starting_address(void)
{
	unsigned int starting_address;
	int i = 1;
	cin >> dec;
	do
	{
		if (i == 1)
		{
			i = 0;
			cout << "请输入操作(十进制)起始地址：";
		}
		else
			cout << "请重新输入操作(十进制)起始地址";
		cin >> starting_address;
		cin.clear();
		cin.sync();
	} while (starting_address > 65535 || starting_address < 0);
	return starting_address;
}

/*********************************************************************************************
* 功能    ：	输入操作数量
* 描述	  ：	操作数量，0x01 1-200  0x03 1-125  0x0xF 1-1968 0x10 1-123
* 输入    ：    function_code 选择的功能码类型
* 输出    ：    operations_Number 对应功能码类型的操作数量
*********************************************************************************************/
unsigned int Input_Operations_number(int function_code)
{
	unsigned int operations_Number;
	int i = 1;
	bool flage = true;
	printf("  功能码  0x01    0x03    0x0F   0x10\n");
	printf("操作数量  1-2000  1-125  1-1968  1-123\n");
	cin >> dec;
	do
	{
		if (i == 1)
		{
			i = 0;
			cout << "请输入操作(十进制)数量：";
		}
		else
			cout << "请重新输入操作(十进制)数量";
		cin >> operations_Number;
		cin.clear();
		cin.sync();
		if (function_code == 0x01 && operations_Number >= 1 && operations_Number <= 65535)
			flage = false;
		if (function_code == 0x03 && operations_Number >= 1 && operations_Number <= 65535)
			flage = false;
		if (function_code == 0x0F && operations_Number >= 1 && operations_Number <= 65535)
			flage = false;
		if (function_code == 0x10 && operations_Number >= 1 && operations_Number <= 65535)
			flage = false;
	} while (flage);
	return operations_Number;
}

/*********************************************************************************************
* 功能    ：	检测写入数据
* 描述	  ：	检测写入数据格式是否正确
* 输入    ：    write_date 写入的数据 write_date_number写入数据的组数
* 输出    ：    true  写入数据格式正确
*				false 写入数据格式错误
*********************************************************************************************/
bool Check_Write_date(string write_date, int write_date_number)
{
	int write_date_len = 2 * write_date_number + write_date_number - 1;
	if (write_date.length() != write_date_len)
		return false;
	int space_number = 0;
	int temp = -1;
	for (unsigned int i = 0; i < write_date.length(); i++)
	{
		if (write_date[i] == 0x20)//空格位置判定
		{
			if (i - temp == 3)
			{
				space_number++;
				temp = i;
				continue;
			}
			else
				return false;
		}
		if ((write_date[i] >= 0x30 && write_date[i] <= 0x39)) // 1-9
			continue;
		else if ((write_date[i] >= 0x41 && write_date[i] <= 0x5A)) //A-Z
			continue;
		else if ((write_date[i] >= 0x61 && write_date[i] <= 0x7A)) //a-z
			continue;
		else
			return false;
	}
	if (space_number != write_date_number - 1)
		return false;
	return true;
}

/*********************************************************************************************
* 功能    ：	计算写入数据的字节数
* 描述	  ：	通过对应的功能码和操作数量计算对应的数据字节数
* 输入    ：    function_code 选择的功能码类型 operations_Number对应功能码类型的操作数量
* 输出    ：    Write_date_number 写入数据的字节数
*********************************************************************************************/
unsigned int Count_Write_date_number(int function_code, unsigned int operations_number)
{
	unsigned int write_date_number = 0;
	if (function_code == 0x01 || function_code == 0x03)
		return write_date_number;
	if (function_code == 0x0F) //读线圈
	{
		write_date_number = function_code / 8;
		if (operations_number % 8)
			write_date_number++;
	}
	if (function_code == 0x10)
		write_date_number = operations_number * 2;
	return write_date_number;
}

/*********************************************************************************************
* 功能    ：	计算请求报文计算预期响应报文字节数
* 描述	  ：	通过对应的功能码和操作数量计算对应的数据字节数
* 输入    ：    Function_code 选择的功能码类型 Operations_Number对应功能码类型的操作数量
* 输出    ：    Respone_len 预期响应的数据字节数
*********************************************************************************************/
unsigned int Count_Respone_Len(int function_code, unsigned int operations_number)
{
	unsigned int respone_len = 0;
	if (function_code == 0x0F || function_code == 0x10)
		return respone_len;
	if (function_code == 0x01) //读线圈
	{
		respone_len = operations_number / 8;
		if (operations_number % 8)
			respone_len++;
	}
	if (function_code == 0x03)
		respone_len = operations_number * 2;
	return respone_len;
}

/*********************************************************************************************
* 功能    ：	输入写入数据
* 描述	  ：	根据写入数量和功能码类型输入相应的数据
* 输入    ：    Function_code 选择的功能码类型 Operations_Number对应功能码类型的操作数量
* 输出    ：    Write_date 写入的数据 
*********************************************************************************************/
string Input_Write_date(int function_code, unsigned int operations_Number)
{
	int i = 1;
	string write_date;
	unsigned int write_date_number = Count_Write_date_number(function_code, operations_Number);
	if (function_code == 0x01 || function_code == 0x03)
		return write_date;//读线圈/寄存器不需要输入写入数据
	
	printf("请输入N组写入数据（2位16进制为一组，空格间隔）例如：00 00 01 00 20 05\n");
	do
	{
		if (i == 1)
		{
			i = 0;
			printf("请输入%d组写入数据 ：", write_date_number);
		}
		else
			printf("请重新输入%d组写入数据 ：", write_date_number);
		getline(cin, write_date);
		cin.clear();
		cin.sync();
	} while (!Check_Write_date(write_date, write_date_number));

	return write_date;
}

/*********************************************************************************************
* 功能    ：	日志记录
* 描述	  ：	记录每次通信的请求和响应报文
* 输入    ：    *Message 消息帧存放的数组   flage = 1 请求报文  flage = 其他 响应报文
*				Message_len 报文数据长度
* 输出    ：    无
*********************************************************************************************/
void Log_Note(UINT8 *Message, int flage, int message_len)
{
	FILE *fp = NULL;
	time_t timep;
	struct tm *p;
	time(&timep);
	p = gmtime(&timep);
	string recv_str = to_string(1900 + p->tm_year) + "-" + to_string(1 + p->tm_mon) + "-" + to_string(p->tm_mday)
		+ " " + to_string(8 + p->tm_hour) + ":" + to_string(p->tm_min) + ":" + to_string(p->tm_sec) + " Recv:";
	string send_str= to_string(1900 + p->tm_year) + "-" + to_string(1 + p->tm_mon) + "-" + to_string(p->tm_mday)
		+ " " + to_string(8 + p->tm_hour) + ":" + to_string(p->tm_min) + ":" + to_string(p->tm_sec) + " Send:";

	/* 打开文件用于读写 */
	if ((fp = fopen("Modbus_TCP_log.txt", "a+")) == NULL)
	{
		printf("打开文件失败");
		return ;
	}
	if (flage == 1)
	{
		fwrite(send_str.c_str(), send_str.length(), 1, fp);
		for (int i = 0; i < message_len; i++)
		{
			fprintf(fp, "%02X ", Message[i]);
		}
		fprintf(fp,"\n");
	}
	else
	{
		fwrite(recv_str.c_str(), recv_str.length(), 1, fp);
		for (int i = 0; i < message_len; i++)
		{
			fprintf(fp, "%02X ", Message[i]);
		}
		fprintf(fp, "\n");
	}
	fclose(fp);
	fp = NULL;
}

/*********************************************************************************************
* 功能    ：	socket版本
* 描述	  ：	启动socket服务
* 输入    ：    无
* 输出    ：    true  启动成功
*				false 启动失败
*********************************************************************************************/
bool InitSocket_Version(void)
{
	WORD sockVersion = MAKEWORD(2, 2);//使用winsocket2.2版本
	WSADATA wsaData;
	if (WSAStartup(sockVersion, &wsaData) != 0)
	{
		return false;
	}
	return true;
}

/*********************************************************************************************
* 功能    ：	初始化客户端
* 描述	  ：	根据终端输入从站IP地址和端口号连接对应的服务器,设置超时时间
* 输入    ：	无
* 输出    ：    ClientSocket 连接成功的套接字
*********************************************************************************************/
SOCKET Init_Client(void)
{
	if (InitSocket_Version() == 0)
	{
		printf("启动Socket服务失败，请检查Socket版本");
		return INVALID_SOCKET;
	}
	string ip_address;
	unsigned int port_number;
	Input_IP(ip_address, &port_number);
	SOCKET clientSocket = socket(AF_INET, SOCK_STREAM, 0);
	sockaddr_in serversock_in;
	serversock_in.sin_addr.S_un.S_addr = inet_addr(ip_address.c_str());
	serversock_in.sin_family = AF_INET;
	serversock_in.sin_port = htons(port_number);
	if (SOCKET_ERROR == connect(clientSocket, (SOCKADDR*)&serversock_in, sizeof(SOCKADDR)))
	{
		cout << "尝试连接TCP从站失败" << endl;
		return INVALID_SOCKET;
	}
	cout << "连接TCP从站成功" << endl;
	TIMEVAL timeout;
	timeout.tv_sec = 200;  //ms
	timeout.tv_usec = 0;  //us
	setsockopt(clientSocket, SOL_SOCKET, SO_RCVTIMEO, (char *)&timeout, sizeof(struct timeval));//设置接收超时时间
	return clientSocket;
}

void Clear_buf(SOCKET clientSocket)
{
	unsigned long bytesToRecv;
	char temp[500];
	do
	{
		ioctlsocket(clientSocket, FIONREAD, &bytesToRecv);
		if (bytesToRecv != 0)//不等于0时进行清理操作
		{
			if (bytesToRecv > 500)
			{
				recv(clientSocket, temp, 500, 0);
			}
			else
				recv(clientSocket, temp, bytesToRecv, 0);
		}
	} while (bytesToRecv != 0);

}

bool Send_date(SOCKET clientSocket, UINT8 *Request_Message, int request_message_len)
{
	Clear_buf(clientSocket); //清理缓冲区
	int status = send(clientSocket, (char*)Request_Message, request_message_len, 0);
	if (status == (-1))
	{
		return false;
	}
	return true;
}

bool Abnormal_Connection(SOCKET *clientSocket)
{
	printf("连接异常，请检查连接状态。\n");
	printf("**************************** Press Enter To Contioun ****************************\n");
	getchar();
	system("cls");
	closesocket(*clientSocket);
	WSACleanup();
	printf("是否重新连接服务器： 1 重新连接  0 关闭本软件\n");
	int flage = 0;
	do
	{
		cin >> flage;
		cin.clear();
		cin.sync();
	} while (!(flage == 0 || flage == 1));
	if (flage == 0)
		return false;
	system("cls");
	do
	{
		*clientSocket = Init_Client();
	} while (INVALID_SOCKET == *clientSocket);
	system("cls");
	return true;
}

int Recv_date(SOCKET clientSocket, UINT8 *Response_Message)
{
	int Response_Message_len = recv(clientSocket, (char*)Response_Message, 600, 0);
	if (Response_Message_len > 0)
	{
		return Response_Message_len;
	}
	return 0;
}

bool Test_Connection_status(SOCKET clientSocket)
{
	TIMEVAL timeout;
	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
	fd_set  reads;
	FD_ZERO(&reads);
	FD_SET(clientSocket, &reads);
	if (select(0, &reads, 0, 0, &timeout))
		return false;
	else
		return true;
}